/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Matrix.h>
#include <Map.h>
#include <Vector.h>

using namespace FD;
using namespace std;

class MatrixToMap;

DECLARE_NODE(MatrixToMap)
/*Node
 *
 * @name MatrixToMap
 * @category HARK:MISC
 * @description Convert Matrix<ObjectRef> to Map<int, ObjectRef>. If the input is Matrix<float>, the output is Map<int, Matrix<float> >. If the input is Matrix<complex<float> >, the output is Map<int, Matrix<complex<float> > >.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Map<int, Matrix<float> >, Vector<float>, Map<int, Matrix<complex<float> > > or Vector<complex<float> >.
 *
 * @parameter_name OUTPUT_TYPE
 * @parameter_type string
 * @parameter_list map_of_matrix:map_of_row_vectors:map_of_column_vectors
 * @parameter_value map_of_row_vectors
 * @parameter_description Type of the output. If map_of_matrix is selected, the key is zero and the value is the input Matrix. If map_of_row_vectors is selected, the keys are the row indices and the values are Vectors extracted the row components. If map_of_column_vectors is selected, the keys are the column indices and the values are Vectors ectracted the column components. [default: map_of_row_vectors]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int KEY_ZERO = 0;
static const int KEY_ROW_INDEX = 1;
static const int KEY_COL_INDEX = 2;

class MatrixToMap : public BufferedNode {
	int inputID;
	int outputID;
	int key;
	bool debug;

public:
	MatrixToMap(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string key_str = object_cast<String>(parameters.get("OUTPUT_TYPE"));
		if (key_str == "map_of_matrix")              key = KEY_ZERO;
		else if (key_str == "map_of_row_vectors")    key = KEY_ROW_INDEX;
		else if (key_str == "map_of_column_vectors") key = KEY_COL_INDEX;
		else {
			cerr << "******* MatrixToMap: Input error" << endl;
			throw new NodeException(NULL, "MatrixToMap: OUTPUT_TYPE should be \"map_of_matrix\" or \"map_of_row_vectors\" or \"map_of_column_vectors\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}
    
	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* MatrixToMap debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Matrix<float>))                cout << "Matrix<float>" << endl;
			else if (typeid(*input) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		ObjectRef orOutputTemp = ObjectRef(new Map<int, ObjectRef>);
		Map<int, ObjectRef> &mpOrOutput = object_cast<Map<int, ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;

		if (typeid(*input) == typeid(Matrix<float>)) {
			const Matrix<float> &mtFltInput = object_cast<Matrix<float> >(input);

			if (key == KEY_ROW_INDEX) {
				for (int row = 0; row < mtFltInput.nrows(); row++) {
					ObjectRef orOutputVal = ObjectRef(new Vector<float>(mtFltInput.ncols()));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputVal);
					for (int col = 0; col < mtFltInput.ncols(); col++) {
						vcFltOutput[col] = mtFltInput(row, col);
					}
					mpOrOutput.insert(make_pair(row, orOutputVal));
				}
			}
			else if (key == KEY_COL_INDEX) {
				for (int col = 0; col < mtFltInput.ncols(); col++) {
					ObjectRef orOutputVal = ObjectRef(new Vector<float>(mtFltInput.nrows()));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputVal);
					for (int row = 0; row < mtFltInput.nrows(); row++) {
						vcFltOutput[row] = mtFltInput(row, col);
					}
					mpOrOutput.insert(make_pair(col, orOutputVal));
				}
			}
			else { // key == KEY_ZERO
				if (mtFltInput.nrows() * mtFltInput.ncols() > 0) {
					mpOrOutput.insert(make_pair(0, input));
				}
			}
		}
		else if (typeid(*input) == typeid(Matrix<complex<float> >)) {
			const Matrix<complex<float> > &mtCmpInput = object_cast<Matrix<complex<float> > >(input);
			
			if (key == KEY_ROW_INDEX) {
				for (int row = 0; row < mtCmpInput.nrows(); row++) {
					ObjectRef orOutputVal = ObjectRef(new Vector<complex<float> >(mtCmpInput.ncols()));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputVal);
					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						vcCmpOutput[col] = mtCmpInput(row, col);
					}
					mpOrOutput.insert(make_pair(row, orOutputVal));
				}
			}
			else if (key == KEY_COL_INDEX) {
				for (int col = 0; col < mtCmpInput.ncols(); col++) {
					ObjectRef orOutputVal = ObjectRef(new Vector<complex<float> >(mtCmpInput.nrows()));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputVal);
					for (int row = 0; row < mtCmpInput.nrows(); row++) {
						vcCmpOutput[row] = mtCmpInput(row, col);
					}
					mpOrOutput.insert(make_pair(col, orOutputVal));
				}
			}
			else { // key == KEY_ZERO
				if (mtCmpInput.nrows() * mtCmpInput.ncols() > 0) {
					mpOrOutput.insert(make_pair(0, input));
				}
			}
		}
		else {
			cerr << "******* MatrixToMap: Input type is unidentified." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MatrixToMap]", __FILE__, __LINE__));
		}

        if (debug == true) cout << "******* MatrixToMap debug point-Finish" << endl;
	}

	NO_ORDER_NODE_SPEEDUP(MatrixToMap);
};
