/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Map.h>
//#include <Complex.h>//<FDComplex.h>flowdesigner-0.9.1-hark

using namespace FD;
using namespace std;

class VectorToMap;

DECLARE_NODE(VectorToMap)
/*Node
 *
 * @name VectorToMap
 * @category HARK:MISC
 * @description Convert Vector<ObjectRef> to Map<int, ObjectRef>. If the input is Vector<float>, the output is Map<int, Vector<float> >. If the input is Vector<complex<float> >, the output is Map<int, Vector<complex<float> > >.
 *
 * @input_name INPUT
 * @input_type Vector<ObjectRef>
 * @input_description Vector<float> or Vector<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Map<int, Vector<float> > or Map<int, Vector<complex<float> > >.
 *
 * @parameter_name OUTPUT_TYPE
 * @parameter_type string
 * @parameter_list map_of_vector:map_of_vectors
 * @parameter_value map_of_vector
 * @parameter_description Type of the output. If map_of_vector is selected, the key is zero and the value is the input Vector. If map_of_vectors is selected, the keys are indices and the values are Vectors with one component in the input Vector. [default: map_of_vector]

 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/
//If map_of_scalars is selected, the keys are indices of vector and the values are each element in the input Vector.

static const int KEY_ZERO = 0;
static const int KEY_INDEX = 1;

class VectorToMap : public BufferedNode {
	int inputID;
	int outputID;
	int key;
	bool debug;
    
public:
	VectorToMap(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string key_str = object_cast<String>(parameters.get("OUTPUT_TYPE"));
		if (key_str == "map_of_vector")       key = KEY_ZERO;
		//else if (key_str == "map_of_scalars") key = KEY_INDEX;
		else if (key_str == "map_of_vectors") key = KEY_INDEX;
		else {
			cerr << "******* VectorToMap: Input error" << endl;
			//throw new NodeException(NULL, "VectorToMap: OUTPUT_TYPE should be \"map_of_vector\" or \"map_of_scalars\".", __FILE__, __LINE__);
			throw new NodeException(NULL, "VectorToMap: OUTPUT_TYPE should be \"map_of_vector\" or \"map_of_vectors\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* VectorToMap debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Vector<float>))                cout << "Vector<float>" << endl;
			else if (typeid(*input) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		ObjectRef orOutputTemp = ObjectRef(new Map<int, ObjectRef>);
		Map<int, ObjectRef> &mpOrOutput = object_cast<Map<int, ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;

		if (typeid(*input) == typeid(Vector<float>)) {
			const Vector<float> &vcFltInput = object_cast<Vector<float> >(input);

			if (key == KEY_INDEX) {
				for (int col = 0; col < vcFltInput.size(); col++) {
					//ObjectRef orOutputVal = ObjectRef(Float::alloc(vcFltInput[col]));
					//mpOrOutput.insert(make_pair(col, orOutputVal));
					ObjectRef orOutputVal = ObjectRef(new Vector<float>(1));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputVal);
					vcFltOutput[0] = vcFltInput[col];
					mpOrOutput.insert(make_pair(col, orOutputVal));
				}
			} 
			else { // key == KEY_ZERO
				if (vcFltInput.size() > 0) {
					mpOrOutput.insert(make_pair(0, input));
				}
			}
		}
		else if (typeid(*input) == typeid(Vector<complex<float> >)) {
			const Vector<complex<float> > &vcCmpInput = object_cast<Vector<complex<float> > >(input);

			if (key == KEY_INDEX) {
				for (int col = 0; col < vcCmpInput.size(); col++) {
					//ObjectRef orOutputVal = ObjectRef(new Complex<float>(vcCmpInput[col]));
					//mpOrOutput.insert(make_pair(col, orOutputVal));
					ObjectRef orOutputVal = ObjectRef(new Vector<complex<float> >(1));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputVal);
					vcCmpOutput[0] = vcCmpInput[col];
					mpOrOutput.insert(make_pair(col, orOutputVal));
				}
			}
			else { // key == KEY_ZERO
				if (vcCmpInput.size() > 0) {
					mpOrOutput.insert(make_pair(0, input));
				}
			}
		}
		else {
			cerr << "******* VectorToMap: Input type is unidentified." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[VectorToMap]", __FILE__, __LINE__));
		}

        if (debug == true) cout << "******* VectorToMap debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(VectorToMap);
};
