/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>

using namespace FD;
using namespace std;

class VectorToMatrix;

DECLARE_NODE(VectorToMatrix)
/*Node
 *
 * @name VectorToMatrix
 * @category HARK:MISC
 * @description Convert Vector<ObjectRef> to Matrix<ObjectRef>. If the input is Vector<float>, the output is Matrix<float>. If the input is Vector<complex<float> >, the output is Matrix<complex<float> >.
 *
 * @input_name INPUT
 * @input_type Vector<ObjectRef>
 * @input_description Vector<float> or Vector<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name OUTPUT_TYPE
 * @parameter_type string
 * @parameter_list row:column
 * @parameter_value row
 * @parameter_description Type of the output. If row is selected, the output is Nx1 Matrix. If column is selected, the output is 1xN Matrix. N is the size of the input Vector. [default: row]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int ROW_MAJOR_ORDER = 0;
static const int COLUMN_MAJOR_ORDER = 1;

class VectorToMatrix : public BufferedNode {
	int inputID;
	int outputID;
    int order;
	bool debug;
    
public:
	VectorToMatrix(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string order_name = object_cast<String>(parameters.get("OUTPUT_TYPE"));
		if (order_name == "row")         order = ROW_MAJOR_ORDER;
		else if (order_name == "column") order = COLUMN_MAJOR_ORDER;
		else {
			cerr << "******* MatrixToVector: Input error" << endl;
			throw new NodeException(NULL, "MatrixToVector: OUTPUT_TYPE should be \"row\" or \"column\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* VectorToMatrix debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Vector<float>))                cout << "Vector<float>" << endl;
			else if (typeid(*input) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Vector<float>)) {
			const Vector<float> &vcFltInput = object_cast<Vector<float> >(input);
			
			if (order == ROW_MAJOR_ORDER) {
				RCPtr<Matrix<float> > output(new Matrix<float>(vcFltInput.size(), 1));	
				for (int row = 0; row < vcFltInput.size(); row++) {
					(*output)(row, 0) = vcFltInput[row];
				}
				out[count] = output;
			}
			else { // order == COLUMN_MAJOR_ORDER
				RCPtr<Matrix<float> > output(new Matrix<float>(1, vcFltInput.size()));	
				for (int col = 0; col < vcFltInput.size(); col++) {
					(*output)(0, col) = vcFltInput[col];
				}
				out[count] = output;
			}
		}
		else if (typeid(*input) == typeid(Vector<complex<float> >)) {
			const Vector<complex<float> > &vcCmpInput = object_cast<Vector<complex<float> > >(input);

			if (order == ROW_MAJOR_ORDER) {
				RCPtr<Matrix<complex<float> > > output(new Matrix<complex<float> >(vcCmpInput.size(), 1));
				for (int row = 0; row < vcCmpInput.size(); row++) {
					(*output)(row, 0) = vcCmpInput[row];
				}
				out[count] = output;	
			}
			else { // order == COLUMN_MAJOR_ORDER
				RCPtr<Matrix<complex<float> > > output(new Matrix<complex<float> >(1, vcCmpInput.size()));
				for (int col = 0; col < vcCmpInput.size(); col++) {
					(*output)(0, col) = vcCmpInput[col];
				}
				out[count] = output;	
			}
		}
		else {
			cerr << "******* VectorToMatrix: Input type is unidentified" << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[VectorToMatrix]", __FILE__, __LINE__));
		}

        if (debug == true ) cout << "******* VectorToMatrix debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(VectorToMatrix);
};
