/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>

using namespace FD;
using namespace std;

class VectorToVector;

DECLARE_NODE(VectorToVector)
/*Node
 *
 * @name VectorToVector
 * @category HARK:MISC
 * @description Convert ObjectRef of Vector<ObjectRef>. If the input is Vector<float>, the output is Vector<complex<float> >. If the input is Vector<complex<float> >, the output is Vector<float>.
 *
 * @input_name INPUT
 * @input_type Vector<ObjectRef>
 * @input_description Vector<float> or Vector<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description Vector<float> or Vector<complex<float> >.
 *
 * @parameter_name METHOD_COMPLEX_TO_FLOAT
 * @parameter_type string
 * @parameter_list magnitude:real:imaginary
 * @parameter_value magnitude
 * @parameter_description Method of converting complex to float. This is used only when INPUT is Vector<complex<float> >. [default: magnitude]
 *
 * @parameter_name METHOD_FLOAT_TO_COMPLEX
 * @parameter_type string
 * @parameter_list zero:hilbert
 * @parameter_value zero
 * @parameter_description Method of converting float to complex. This is used only when INPUT is Vector<float>. [default: zero]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int COMPLEX_ZERO = 0;
static const int COMPLEX_HILBERT = 1;

static const int SOURCE_MAGNITUDE = 0;
static const int SOURCE_REAL = 1;
static const int SOURCE_IMAGINARY = 2;

class VectorToVector : public BufferedNode {
	int inputID;
	int outputID;
	int method;
	int complex_method;
	bool debug;
    
public:
	VectorToVector(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_str = object_cast<String>(parameters.get("METHOD_COMPLEX_TO_FLOAT"));
		if (method_str == "magnitude")      method = SOURCE_MAGNITUDE;
		else if (method_str == "real")      method = SOURCE_REAL;
		else if (method_str == "imaginary") method = SOURCE_IMAGINARY;
		else {
			cerr << "******* VectorToVector: Input error" << endl;
			throw new NodeException(NULL, "VectorToVector: METHOD_COMPLEX_TO_FLOAT be \"magnitude\" or \"real\" or \"imaginary\".", __FILE__, __LINE__);
		}
		string complex_str = object_cast<String>(parameters.get("METHOD_FLOAT_TO_COMPLEX"));
		if (complex_str == "zero")         complex_method = COMPLEX_ZERO;
		else if (complex_str == "hilbert") complex_method = COMPLEX_HILBERT;
		else {
			cerr << "******* VectorToVector: Input error" << endl;
			throw new NodeException(NULL, "VectorToVector: METHOD_FLOAT_TO_COMPLEX should be \"zero\" or \"hilbert\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* VectorToVector debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Vector<float>))                cout << "Vector<float>" << endl;
			else if (typeid(*input) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Vector<float>)) {
			const Vector<float> &vcFltInput = object_cast<Vector<float> >(input);

			ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(vcFltInput.size()));
			Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
			out[count] = orOutputTemp;

			if (complex_method == COMPLEX_HILBERT) {
				for (int col = 0; col < vcFltInput.size(); col++) {
					vcCmpOutput[col] = complex<float>(vcFltInput[col], abs(vcFltInput[col]));
				}
			}
			else { // complex_method == COMPLEX_ZERO
				for (int col = 0; col < vcFltInput.size(); col++) {
					vcCmpOutput[col] = complex<float>(vcFltInput[col], 0.0f);
				}
			}
		}
		else if (typeid(*input) == typeid(Vector<complex<float> >)) {
			const Vector<complex<float> > &vcCmpInput = object_cast<Vector<complex<float> > >(input);

			ObjectRef orOutputTemp = ObjectRef(new Vector<float>(vcCmpInput.size()));
			Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
			out[count] = orOutputTemp;

			if (method == SOURCE_REAL) {
				for (int col = 0; col < vcCmpInput.size(); col++) {
					vcFltOutput[col] = vcCmpInput[col].real();
				}
			}
			else if (method == SOURCE_IMAGINARY) {
				for (int col = 0; col < vcCmpInput.size(); col++) {
					vcFltOutput[col] = vcCmpInput[col].imag();
				}
			}
			else { // method == SOURCE_MAGNITUDE
				for (int col = 0; col < vcCmpInput.size(); col++) {
					vcFltOutput[col] = abs(vcCmpInput[col]);
				}
			}
		}
		else {
			cerr << "******* VectorToVector: Input type is unidentified" << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[VectorToVector]", __FILE__, __LINE__));
		}

        if (debug == true ) cout << "******* VectorToVector debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(VectorToVector);
};
